
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Advanced size solver
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%              by Maximiliano Cristia' and  Gianfranco Rossi
%                              January 2020
%
%                           Revised May 2020
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%size_solver_on.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%     size  solver      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%% SAT solver  (by Howe & King) %%%%%

initialise(_).  

search(Clauses, Vars, Sat, _) :-
    sat(Clauses, Vars),
    !,
    Sat = true.
search(_Clauses, _Vars, false, _).

sat(Clauses, Vars) :-
    problem_setup(Clauses), elim_var(Vars).

elim_var([]). 
elim_var([Var | Vars]) :-
    elim_var(Vars), (Var = true; Var = false). 

problem_setup([]). 

problem_setup([Clause | Clauses]) :- 
    clause_setup(Clause),
    problem_setup(Clauses). 

clause_setup([Pol-Var | Pairs]) :-
    set_watch(Pairs, Var, Pol). 

set_watch([], Var, Pol) :- Var = Pol. 
set_watch([Pol2-Var2 | Pairs], Var1, Pol1) :- 
    freeze(Var1,V=u), %u is simply a flag
    freeze(Var2,V=u),
    freeze(V, watch(Var1,Pol1,Var2,Pol2,Pairs)).

watch(Var1, Pol1, Var2, Pol2, Pairs) :- 
    nonvar(Var1) ->
      update_watch(Var1, Pol1, Var2, Pol2, Pairs);
      update_watch(Var2, Pol2, Var1, Pol1, Pairs). 

update_watch(Var1, Pol1, Var2, Pol2, Pairs) :-
    Var1 == Pol1 -> true; set_watch(Pairs, Var2, Pol2). 

%%%%%%%% construct and solve the boolean formula %%%%%

%pi(+F,+Vars,+Neqs,-AdmSol)
pi(F,Vars,Neqs,AdmSol) :-
    bool_code(F,Vars,L),
    %DBG write('Boolean formula: \n\t'), write(L),nl,%get(_),
    (setof(Vars,sat(L,Vars),Arrs),! ; Arrs=[]),
    %DBG write('Boolean solutions: \n\t'),write(Arrs),nl,get(_),
    powerset(Arrs,AdmSol),    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%AdmSol \== [],   %10.4.2020
    %DBG write('Admissible solution: \n\t'),write(AdmSol),nl,get(_),
    check_neq(Neqs,Vars,AdmSol).
       
%bool_code(+F,+Vars,-L) 
bool_code(F,Vars,L) :-
    bool_code_formula(F,Vars,FL),
    one_true(Vars,VL),
    append(FL,[VL],L).     
       
bool_code_formula(true,_,[]) :- !. 
bool_code_formula((F1 & F),Vars,L) :-
    term_variables(F1,F1Vars),
    setlog:subset_strong(F1Vars,Vars),
    bool_code_constraint(F1,L1),!,
    bool_code_formula(F,Vars,L2),
    append(L1,L2,L).
bool_code_formula((_F1 & F),Vars,L) :-
    bool_code_formula(F,Vars,L).
               
bool_code_constraint(un(X,Y,Z),F) :- !,
    var(X),var(Y),var(Z),          
    var_or_false([X,Y,Z],[A2,B2,C2]),
    F=[[false-C2,true-B2,true-A2],[false-A2,true-C2],[true-C2,false-B2]]. 
bool_code_constraint(disj(X,Y),F) :- !,
    var(X),var(Y),       
    var_or_false([X,Y],[A2,B2]),
    F=[[false-A2,false-B2]].
bool_code_constraint(inters(X,Y,Z),F) :- !,
    var(X),var(Y),var(Z),         
    var_or_false([X,Y,Z],[A2,B2,C2]),
    F=[[false-C2,true-A2],[false-C2,true-B2],[false-A2,false-B2,true-C2]].
bool_code_constraint(subset(X,Y),F) :- !,
    var(X),var(Y),        
    var_or_false([X,Y],[A2,B2]),
    F=[[false-A2,true-B2]].     
bool_code_constraint(diff(X,Y,Z),F) :- !,
    var(X),var(Y),var(Z),         
    var_or_false([X,Y,Z],[A2,B2,C2]),
    F=[[false-C2,true-A2],[false-C2,false-B2],[false-A2,true-B2,true-C2]].

one_true([],[]).
one_true([V|VR],[true-V|VRtrue]) :-
    one_true(VR,VRtrue).

var_or_false([],[]).
var_or_false([X|RX],[V|RV]) :-
    (var(X) -> V=X ; V=false),
    var_or_false(RX,RV).

%check_neq(+Neqs,+Vars,+Sol).
check_neq([],_Vars,_Sol) :- !.
check_neq(_Neqs,_Vars,[]) :- !.    %10.4.2020
check_neq([A neq B|Neqs],Vars,Sol) :- 
    var(A),var(B),
    member_th(A,Vars,K1),        % search variable A in Vars and returns its index K1
    member_th(B,Vars,K2),!,      % search variable B in Vars and returns its index K2
    length(Vars,NVars),  ck_neq_sol(K1,K2,NVars,Sol), 
    check_neq(Neqs,Vars,Sol) .   
check_neq([A neq B|Neqs],Vars,Sol) :- 
    var(A), B=={}, 
    member_th(A,Vars,K1),!,        
    length(Vars,NVars), ck_neq_sol(K1,NVars,Sol), 
    check_neq(Neqs,Vars,Sol).   
check_neq([B neq A|Neqs],Vars,Sol) :-   
    var(A), B=={},  
    member_th(A,Vars,K1),!,  
    length(Vars,NVars), ck_neq_sol(K1,NVars,Sol), 
    check_neq(Neqs,Vars,Sol).   
check_neq([A neq B|_Neqs],_Vars,_Sol) :-   
    A=={}, B=={},!, fail.      
check_neq([_A neq _B|Neqs],Vars,Sol) :- 
    check_neq(Neqs,Vars,Sol).   

ck_neq_sol(K1,K2,NVars,Sol) :-
    length(NeqSol,NVars), nth1(K1,NeqSol,true), nth1(K2,NeqSol,false),
    member(NeqSol,Sol),!.
ck_neq_sol(K1,K2,NVars,Sol) :-
    length(NeqSol,NVars), nth1(K1,NeqSol,false), nth1(K2,NeqSol,true),
    member(NeqSol,Sol).

ck_neq_sol(K1,NVars,Sol) :-
    length(NeqSol,NVars), nth1(K1,NeqSol,true),
    %DBG write('NeqSol: '),write(NeqSol),nl,
    %DBG write(member(3,NeqSol,Sol)),nl,
    member(NeqSol,Sol),
    !.

powerset([],[]).
powerset([_H|T],P) :- powerset(T,P).
powerset([H|T],[H|P]) :- powerset(T,P).


%%%%%%%% construct and solve the arithmetic formula %%%%%

mk_arithm_formula(F,SetVars,Sols,R) :-
    length(Sols,NSols),
    length(ViVars,NSols),
    replace_size(F,SetVars,ViVars,Sols,R).

replace_size(true,_,ViVars,_,R) :- !,
    all_positive(ViVars,R).
replace_size((size(S,N) & F2),SetVars,ViVars,Sols,[N is 0|R] ) :-       
    nonvar(S), S={},!,   
    replace_size(F2,SetVars,ViVars,Sols,R).
replace_size((size(S,N) & F2),SetVars,ViVars,Sols,[N is Expr|R] ) :- 
    member_th(S,SetVars,K),!,     % search variable S in SetVars and returns its index K
    mk_expr(K,Sols,ViVars,Expr),
    replace_size(F2,SetVars,ViVars,Sols,R).
replace_size((F1 & F2),SetVars,ViVars,Sols,[F1|R]) :- 
    F1 = (A neq _B),
    attvar(A),!,   
    replace_size(F2,SetVars,ViVars,Sols,R).
replace_size((F1 & F2),SetVars,ViVars,Sols,[F1|R]) :- 
    F1 =.. [Op,_,_],
    member(Op,[is,=<,<,>=,>,=]),!,   
    replace_size(F2,SetVars,ViVars,Sols,R).
replace_size((_F1 & F2),SetVars,ViVars,Sols,R) :- 
    replace_size(F2,SetVars,ViVars,Sols,R).

mk_expr(_,[],_,0) :-!.
mk_expr(K,[Sol|Sols],[V|Vars],Expr)  :- !,
    nth1(K,Sol,B),           %get the k-th element in Sol and return its value (true/false) in B 
    (B==true,!,Expr = 1*V + Expr1
     ;
     B==false,Expr =  0*V + Expr1
    ),
    mk_expr(K,Sols,Vars,Expr1) .     

member_th(X,[Y|_],1) :-
    X==Y,!.
member_th(X,[_Y|R],N) :-
    member_th(X,R,M),
    N is M + 1.

all_positive([],[]) :- !.
all_positive([V|Vars],[V > 0|R]) :-
    all_positive(Vars,R). 

%%%%%%%% find set variables %%%%%
     
find_setvars(F,Vars,Neqs) :-
    find_size_neq(F,F,[],SizeVars,NewF,Neqs),
    find_vars(NewF,SizeVars,Vars).
       
%find_size_neq(+F,+FAll,+Vars,-NewVars,-NewF,-Neqs)    (new parameter FAll = the whole formula
find_size_neq(true,_,Vars,Vars,true,[]) :- !.
find_size_neq((size(S,_) & F),FAll,Vars,NewVars,NewF,Neqs) :- 
    var(S),        
    setlog:member_strong(S,Vars), !,
    find_size_neq(F,FAll,Vars,NewVars,NewF,Neqs).
find_size_neq((size(S,_) & F),FAll,Vars,NewVars,NewF,Neqs) :- 
    var(S),
    member_constraint(S,FAll,V1,_),V1\==[],!,         
    find_size_neq(F,FAll,[S|Vars],NewVars,NewF,Neqs).
find_size_neq((A neq B & F),FAll,Vars,NewVars,(A neq B & NewF),[A neq B|Neqs]) :- !,
    find_size_neq(F,FAll,Vars,NewVars,NewF,Neqs).
find_size_neq((F1 & F),FAll,Vars,NewVars,(F1 & NewF),Neqs) :- !,
    find_size_neq(F,FAll,Vars,NewVars,NewF,Neqs).

find_vars(_F,[],[]) :- !.  
find_vars(true,Vars,Vars) :- !.   
find_vars(F,[V|Vars],NewVars) :-
    member_constraint(V,F,V1,F1),
    V1 \== [],!,                      
    cond_append(V1,Vars,Vars1),
    find_vars(F1,Vars1,NewVars1),
    cond_append(Vars1,NewVars1,NewVars).
find_vars(F,[_V|Vars],NewVars) :- 
    find_vars(F,Vars,NewVars).

%member_constraint(+V,+F,-NewV,-NewF)     %find all set constraints in F containing variable V and
%                                         %return the new variables (NewV) and the formula without the found constraints  
member_constraint(_V,true,[],true) :- !.      
member_constraint(V,(un(X,Y,Z) & true),[X,Y,Z],true) :-    
    one_of(V,X,Y,Z),!.
member_constraint(V,(un(X,Y,Z) & F),[X,Y,Z|NewV],NewF) :- 
    one_of(V,X,Y,Z),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(disj(X,Y) & true),[X,Y],true) :- 
    one_of(V,X,Y),!.
member_constraint(V,(disj(X,Y) & F),[X,Y|NewV],NewF) :- 
    one_of(V,X,Y),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(subset(X,Y) & true),[X,Y],true) :- 
    one_of(V,X,Y),!.
member_constraint(V,(subset(X,Y) & F),[X,Y|NewV],NewF) :- 
    one_of(V,X,Y),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(inters(X,Y,Z) & true),[X,Y,Z],true) :- 
    one_of(V,X,Y,Z),!.   
member_constraint(V,(inters(X,Y,Z) & F),[X,Y,Z|NewV],NewF) :- 
    one_of(V,X,Y,Z),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(diff(X,Y,Z) & true),[X,Y,Z],true) :- 
    one_of(V,X,Y,Z),!.    
member_constraint(V,(diff(X,Y,Z) & F),[X,Y,Z|NewV],NewF) :- 
    one_of(V,X,Y,Z),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(X neq Y & true),[X,Y],true) :- 
    var(Y),!,
    (V==X,! ; V==Y).
member_constraint(V,(X neq Y & F),[X,Y|NewV],NewF) :- 
    var(Y),
    one_of(V,X,Y),!,
    member_constraint(V,F,NewV,NewF).
member_constraint(V,(F1 & F),NewV,(F1 & NewF)) :- 
    member_constraint(V,F,NewV,NewF).  

one_of(V,X,_Y,_Z) :- V==X,!.     
one_of(V,_X,Y,_Z) :- V==Y,!.
one_of(V,_X,_Y,Z) :- V==Z.

one_of(V,X,_Y) :- V==X,!.        
one_of(V,_X,Y) :- V==Y.

%cond_append(+L1,+L2,-L3)   
cond_append([],Vars,Vars).
cond_append([V|R],Vars,VarsNew) :-
    cond_append1(V,Vars,Vars1),
    cond_append(R,Vars1,VarsNew).     

cond_append1(V,Vars,Vars) :-
    setlog:member_strong(V,Vars),!.
cond_append1(V,Vars,[V|Vars]).


%%%%%%%% size_solve %%%%%

size_solve(true) :- !.
size_solve(F) :-
    %DBG write('Input formula: \n\t'),write(F),nl,
    find_setvars(F,RVars,Neqs), 
    reverse(RVars,Vars),
    %DBG write('Set variables: '),write(Vars),nl,
    size_solve(F,Vars,Neqs).

size_solve(F,SetVars,Neqs) :-
    (SetVars\==[] ->                      
        pi(F,SetVars,Neqs,Sols),
    	%DBG write('Subset of boolean solutions: \n\t'),write(Sols),nl,
    	mk_arithm_formula(F,SetVars,Sols,ArithmL),     
    	%DBG  write('Arithmetic formula: \n\t'),write(ArithmL),nl,
    	solve_Q_all(ArithmL),
    	%DBG write('Arithmetic formula: \n\t'),write(ArithmL),nl,
    	term_variables(ArithmL,IntVars),  
    	%DBG write('Arithmetic variables: \n\t'),write(IntVars),nl,
    	bb_inf(IntVars,1,_,_)
    ; 
        true
    ).
     

%%%%%%%% check size constraints in the constraint list CList %%%%%

cond_check_size(RedC,Nsize) :-
    %DBG write('number of size constraints: '),write(Nsize),nl,
    (Nsize >= 1 ->        %KKKKKKKKKKKKKKKK
        b_getval(int_solver,CurrentSlv),        
        b_setval(int_solver,clpq),
        check_size(RedC),!,
        b_setval(int_solver,CurrentSlv)
    ;
        true
    ).

check_size(CList) :-    
    %DBG write('\n****** Input formula: '),nl,write(CList),nl,  
    get_formula(CList,Formula),
    %DBG write('****** Simplified formula: '),nl,write(Formula),nl,   
    copy_term(Formula,FormulaNew,_),
    %DBG write('****** Simplified formula copied: '),nl,write(FormulaNew),nl,%get(_),   
    size_solve(FormulaNew).

get_formula([],true) :-!.             
get_formula([glb_state(C)|R],C & D) :- !,  %add arithmetic constraints (stored in glb_state terms)
    get_formula(R,D).
get_formula([solved(C,_,_,_)|R],CD) :- !,  
    get_formula([C|R],CD).
get_formula([size(S1,S2)|R],size(S1,S2) & D) :-
     var(S1),!,
     get_formula(R,D).
get_formula([C|R],C & D) :-                %add set constraints in solved form
     solved_set_constraint(C),!,
     get_formula(R,D).
get_formula([X neq Y|R],C & D) :-    
     attvar(X),!, 
     (C = (X > Y)
      ;
      C = (X < Y)
     ), 
     get_formula(R,D).
get_formula([subset(S,I)|R],CD) :-  
     var(S), 
     nonvar(I), I=int(A,B),!,  
     ( 
     C=(subset(S,_) & M is B-A+1 & M > 0 & size(S,N) & N >= 0 & N =< M & true)  %KKKKKKKK
      ;
      %S={},
      %C=(M is B-A+1 & M =< 0 & true)
      C=(subset(S,_) & M is B-A+1 & M =< 0 & size(S,N) & N >= 0 & N =< 0 & true)
     )
     , 
     get_formula(R,D),
     conj_append(C,D,CD).
get_formula([_C|R],D) :-
     get_formula(R,D).

solved_set_constraint(un(X,Y,Z)) :- !,
     var(X),var(Y),var(Z).
solved_set_constraint(subset(X,Y)) :- !,
     var(X),var(Y).
solved_set_constraint(disj(X,Y)) :- !,
     var(X),var(Y).
solved_set_constraint(inters(X,Y,Z)) :- !,
     var(X),var(Y),var(Z).
solved_set_constraint(diff(X,Y,Z)) :- !,
     var(X),var(Y),var(Z).

%%%%% Extra predicates for clpq and clpfd

solve_Q_all([]).        
solve_Q_all([C|ConstrList]) :-             % solve the constraint 'Constr' using the CLP(Q) solver
    solve_Q(C,_),
    solve_Q_all(ConstrList).



